{% extends "base.html" %}

{% block main %}
{% include "_title.html" %}

<div class="box">
  <h2 class="subtitle">Providers</h2>

  <div class="buttons">
  {% for provider in providers %}
    <a class="button is-medium is-fullwidth" href="{{ provider.url }}" target="_blank" rel="noopener noreferrer">
      <figure class="icon">
        <img class="icon" src="{{ provider.logo }}" />
      </figure>
      <label>{{ provider.label }}</label>
    </a>
  {% endfor %}
  </div>
</div>
{% endblock %}
