(function(){"use strict";var e={5342:function(e,t,n){n(6992),n(8674),n(9601),n(7727);var r=n(8345),o=n(144),i=n(594),a=function(){var e=this,t=e._self._c;return t("div",{attrs:{id:"app"}},[t("router-view")],1)},u=[],s={name:"Embed"},l=s,c=n(1001),f=(0,c.Z)(l,a,u,!1,null,null,null),d=f.exports,p=n(6790),m=n(6080),h=function(){var e=this,t=e._self._c;return t("router-view-layout",[e.isLoading?t("div",{staticClass:"box is-marginless"},[t("progress",{staticClass:"progress is-small is-info"})]):t("div",{staticClass:"box is-marginless"},[t("div",{staticClass:"content has-text-centered"},[t("p",{staticClass:"is-italic"},[e._v(e._s(e.error))])])]),t("div",{staticClass:"is-pulled-right mt-05r scale-08"},[t("a",{staticClass:"is-size-7",attrs:{href:"https://meltano.com",target:"_blank"}},[t("span",{staticClass:"is-inline-block has-text-grey"},[e._v("Made with")]),t("Logo",{staticClass:"ml-05r"})],1)])])},v=[],b=(n(1539),n(9240)),y={generate:function(e){return i.Z.post(b.Z.apiUrl("embeds","embed"),e)},load:function(e,t){var n=b.Z.apiUrl("embeds/embed",e);return t&&(n+="?today=".concat(t)),i.Z.get(n)}},g=n(9964),w=n(4309),O={name:"ResourceEmbed",components:{Logo:g.Z,RouterViewLayout:w.Z},props:{token:{type:String,default:null},today:{type:String,default:null}},data:function(){return{error:null,isLoading:!0,resource:null,resourceType:null}},created:function(){this.initialize()},methods:{initialize:function(){var e=this;y.load(this.token,this.today).then((function(t){e.resource=t.data.resource,e.resourceType=t.data.resourceType})).catch((function(t){e.error=t.response.data.code})).finally((function(){return e.isLoading=!1}))}}},Z=O,_=(0,c.Z)(Z,h,v,!1,null,null,null),C=_.exports,k=new r.ZP({mode:"history",routes:[{path:"/-/embed/:token",name:"resource-embed",component:C,meta:{title:"Meltano Resource Embed"},props:function(e){return(0,m.Z)((0,m.Z)({},e.params),{},{today:e.query.today})}}]}),x=k;o["default"].use(r.ZP),i.Z.defaults.headers.common["X-JSON-SCHEME"]="camel",o["default"].prototype.$flask=(0,p.Z)(),new o["default"]({el:"#app",router:x,render:function(e){return e(d)}})}},t={};function n(r){var o=t[r];if(void 0!==o)return o.exports;var i=t[r]={id:r,loaded:!1,exports:{}};return e[r].call(i.exports,i,i.exports,n),i.loaded=!0,i.exports}n.m=e,function(){var e=[];n.O=function(t,r,o,i){if(!r){var a=1/0;for(c=0;c<e.length;c++){r=e[c][0],o=e[c][1],i=e[c][2];for(var u=!0,s=0;s<r.length;s++)(!1&i||a>=i)&&Object.keys(n.O).every((function(e){return n.O[e](r[s])}))?r.splice(s--,1):(u=!1,i<a&&(a=i));if(u){e.splice(c--,1);var l=o();void 0!==l&&(t=l)}}return t}i=i||0;for(var c=e.length;c>0&&e[c-1][2]>i;c--)e[c]=e[c-1];e[c]=[r,o,i]}}(),function(){n.n=function(e){var t=e&&e.__esModule?function(){return e["default"]}:function(){return e};return n.d(t,{a:t}),t}}(),function(){n.d=function(e,t){for(var r in t)n.o(t,r)&&!n.o(e,r)&&Object.defineProperty(e,r,{enumerable:!0,get:t[r]})}}(),function(){n.g=function(){if("object"===typeof globalThis)return globalThis;try{return this||new Function("return this")()}catch(e){if("object"===typeof window)return window}}()}(),function(){n.o=function(e,t){return Object.prototype.hasOwnProperty.call(e,t)}}(),function(){n.r=function(e){"undefined"!==typeof Symbol&&Symbol.toStringTag&&Object.defineProperty(e,Symbol.toStringTag,{value:"Module"}),Object.defineProperty(e,"__esModule",{value:!0})}}(),function(){n.nmd=function(e){return e.paths=[],e.children||(e.children=[]),e}}(),function(){n.j=991}(),function(){var e={991:0};n.O.j=function(t){return 0===e[t]};var t=function(t,r){var o,i,a=r[0],u=r[1],s=r[2],l=0;if(a.some((function(t){return 0!==e[t]}))){for(o in u)n.o(u,o)&&(n.m[o]=u[o]);if(s)var c=s(n)}for(t&&t(r);l<a.length;l++)i=a[l],n.o(e,i)&&e[i]&&e[i][0](),e[i]=0;return n.O(c)},r=self["webpackChunkmeltano_webapp"]=self["webpackChunkmeltano_webapp"]||[];r.forEach(t.bind(null,0)),r.push=t.bind(null,r.push.bind(r))}();var r=n.O(void 0,[998,64],(function(){return n(5342)}));r=n.O(r)})();
//# sourceMappingURL=embed-legacy.062c14d1.js.map